//!\file File.h File name converter and file management functions
//Luke Lenhart (2007)
//See /docs/License.txt for details on how this code may be used.

#pragma once
#include <string>
#include <vector>

namespace MPMA
{
    //!\brief Construct this with a non-absolute filename using either windows or linux style, to
    //! generate a filename for whatever platform this is running on.
    //!Filenames that begin with ~ are relative to the users home directory.
    class Filename
    {
    public:
        Filename(const std::string &filename); //!<ctor
        inline Filename(const char *filename) { *this=Filename(std::string(filename)); } //!<ctor
        
        inline operator const std::string&() const { return fname; } //!<Gets a string that represents the filename.
        
        inline const char* c_str() const { return fname.c_str(); } //!<Gets a c string that represents the filename.
        
    private:
        std::string fname;
        
        //remove rediculousness from a filename
        void PurifyFilename();
    };

    //!Represents a directory that can be browsed and navigated.
    class Directory
    {
    public:
        Directory(); //!<ctor. defaults to the current directory
        Directory(const Filename &initialPath); //!<ctor.  specify "~" to start in the users home directory
        
        //!Create a directory if it does not already exist.  Returns true if the directory existed or was created.
        bool CreateDirectory();
        
        //!Retrieves a list of files in this directory.
        std::vector<std::string> GetFiles() const;
        //!Retrieves a list of subdirectories in this directory.
        std::vector<std::string> GetSubDirectories() const;
        
        //!Changes the directory to one tier above the current directory.
        void StepOut();
        //!Steps into a subdirectory of the current directory.  It does not have to exist.
        void StepInto(const std::string &subdir);
        
        //!Retrieves a string that represents the current path.
        inline operator const std::string&() const {return path;}
        
        //!Retrieves a Filename for a file within the current path.
        inline Filename File(const std::string &fname) const {return Filename(path+"/"+fname);}
    
    private:
        std::string path;
    };
    
} //namespace MPMA
