//Luke Lenhart, 2001-2007
//See /docs/License.txt for details on how this code may be used.

#include "MemMan.h"
#include "Debug.h"
#include "DebugRouter.h"
#include "MiscStuff.h"
#include "Locks.h"
#include "Vary.h"

using MPMA::Vary;

#ifdef MEM_TRACK_OLD_FREE
    #include <vector>
    namespace
    {
        std::vector<MemMan::SAlloc> oldFreedAllocs;
    }
#endif

MemMan mMan;

//local hrelpers
namespace
{
#ifdef MEMMAN_TRACING //if management is enabled

    class AutoInitMemMan
    {
    public:
        //hookup init callbacks
        AutoInitMemMan()
        {
            MPMA::Internal_AddInitCallback(MemManInitialize,-9999);
            MPMA::Internal_AddShutdownCallback(MemManShutdown,-9999);
        }
    private:
        static void MemManInitialize()
        {
            mMan.InternalInitialize();
        }
        static void MemManShutdown()
        {
            mMan.InternalShutdown();
        }
    } autoInitMemMan;

    const uint8 PAD_BYTE=0xA5; // 1010 0101

    //sets the padding around an allocation
    void SetAllocPadding(uint8 *mem, uint objSize)
    {
        memset(mem,PAD_BYTE,dbgN2_padding);
        memset(mem+objSize+dbgN2_padding,PAD_BYTE,dbgN2_padding);
    }

    //checks that an allocations padding has not been corrupted
    bool VerifyAllocPadding(uint8 *mem, uint objSize)
    {
        bool bad=false;

        //check mem before
        try
        {
            for (int i=0;i<dbgN2_padding;++i)
            {
                if (*(mem+i) != PAD_BYTE)
                {
                    MPMA::ErrorReport()<<"WARNING: Damage to memory BEFORE an allocation.\n";
                    bad=true;
                    break;
                }
            }
        }
        catch(...)
        {
            MPMA::ErrorReport()<<"WARNING: CRASH trying to verify the padding BEFORE an allocation.\n";
            bad=true;
        }

        //check mem after
        try
        {
            for (int i=0;i<dbgN2_padding;++i)
            {
                if (*(mem+objSize+dbgN2_padding+i) != PAD_BYTE)
                {
                    MPMA::ErrorReport()<<"WARNING: Damage to memory AFTER an allocation.\n";
                    bad=true;
                    break;
                }
            }
        }
        catch(...)
        {
            MPMA::ErrorReport()<<"WARNING: CRASH trying to verify the padding AFTER an allocation.\n";
            bad=true;
        }

        return bad;
    }

#endif //tracing
};


//
MemMan::MemMan()
{
    m_ready=false;
}

MemMan::~MemMan()
{
    m_ready=false;
}

//setup initial memman state
void MemMan::InternalInitialize()
{
#ifdef MEMMAN_TRACING //if management is enabled
    m_critErrors=false;
    m_head=0;
    maxMem=0;
    curHintAllocID=1;
#endif
    syncLock=0;
    syncLock=new MPMA::MutexLock;
    m_ready=true;
}

//check for leaks, then shut down memman
void MemMan::InternalShutdown()
{
    //the ErrorReport has already been shut down at this point.. hence we need our own callback
#ifdef MEMMAN_TRACING //if management is enabled
    CheckForPaddingCorruption();
#endif

    m_ready=false;

#ifdef MEMMAN_TRACING //if management is enabled

    //spit out max memory usage
#ifdef TRACK_MAX_MEM
    ReportLeak("Maximum system memory used at any given time (from tracable sources): ");
    if (maxMem>1024*1024*4) ReportLeak(Vary(maxMem/(1024*1024))+"MB\n\n"); //if >4MB then show in MB
    else if (maxMem>1024*16) ReportLeak(Vary(maxMem/(1024))+"KB\n\n"); //if >16KB then show in KB
    else ReportLeak(Vary(maxMem)+" bytes\n\n"); //else just show in bytes
#endif

    //check for critical errors
    if (m_critErrors) //tell them something very bad happened
    {
        ReportLeak("Critical memory management errors occured.\n\n");
    }

    //check if they didn't free memory
    if (m_head)
    {
        SAlloc *cur;

        //lets see how much was leaked total...
        size_t leakedBytes=0;
        int leakedCount=0;
        for (cur=m_head;cur;cur=cur->next)
        {
            leakedBytes+=cur->size;
            leakedCount++;
        }

        ReportLeak("\nTotal memory leaks: "+Vary((int)leakedBytes)+" bytes in "+Vary(leakedCount)+" allocation(s).\n");

        //report all leaks
        ReportLeak("\n -- List of memory leaks --\n");
        for (cur=m_head;cur;)
        {
            //report
            std::string leak;
            if (cur->type==EMAT_One) leak+="Object \"";
            else leak+="Array of \"";
            leak+=cur->objName+"\" ("+Vary((int)cur->size)+" bytes total):\n";
            leak+="  -Allocated in "+cur->srcFile+" line "+Vary(cur->srcLine)+".\n";
            leak+="  -call stack of allocation:\n"+cur->allocStack+"\n";

            ReportLeak(leak);

            //free this object and advance cur
            SAlloc* next=cur->next;
            delete cur;
            cur=next;
        }
    }

    //check if they didn't free hints
    if (hints.size()>0)
    {
        ReportLeak("Warning: the following memory hints have not been freed:\n");

        for (std::list<Hint>::iterator i=hints.begin(); i!=hints.end(); ++i)
        {
            ReportLeak(Vary((*i).bytes)+" bytes allocated at:\n"+(*i).stack+"\n");
        }

        ReportLeak("\n");
    }

#endif

    //our stuff
    delete syncLock;
    syncLock=0;
}

//adds a callback function to call to report memory leaks to
void MemMan::AddLeakReportCallback(void (*leakCallback)(const std::string&))
{
#ifdef MEMMAN_TRACING //if management is enabled
    leakCallbacks.push_back(leakCallback);
#endif
}

#ifdef MEMMAN_TRACING //if management is enabled

//
void MemMan::ReportLeak(const std::string &leak)
{
    for (std::list<void (*)(const std::string &)>::iterator i=leakCallbacks.begin(); i!=leakCallbacks.end(); ++i)
        (*i)(leak);
}

//records allocation of memory
void MemMan::TraceAlloc(EMemAllocType type, volatile void *addr, void *allocMem, uint size, const char *name, const char *file, int line, int count)
{
    if (!m_ready)
    {
        MPMA::ErrorReport()<<"MemMan being used before it is ready!\n";
        BREAKPOINT;
        return;
    }

    if (!addr)
    {
        MPMA::ErrorReport()<<"Out of memory trying to alloc?  A call to new must have failed...\n";
        MPMA::ErrorReport()<<"  -call stack:\n"<<MPMA::GetCallStack()<<"\n";
        return;
    }
    
    MPMA::TakeMutexLock autoSync(*syncLock);

    //create allocation data
    SAlloc *newAlloc=new SAlloc;
    if (!newAlloc)
    {
        MPMA::ErrorReport()<<"Alloc node in MemMan to trace an allocation with failed to allocated.  Out of memory...?\n";
        return;
    }

    newAlloc->type=type;
    newAlloc->addr=(void*)addr;
    newAlloc->allocMem=allocMem;
    newAlloc->size=size;
    newAlloc->count=count;
    newAlloc->objName=name;
    newAlloc->srcFile=file;
    newAlloc->srcLine=line;
#ifdef SAVE_ALLOC_CALL_STACK
    newAlloc->allocStack=MISC::StripFirstLine(MPMA::GetCallStack());
#endif

    //add it to front list
    newAlloc->next=m_head;
    m_head=newAlloc;

    //pad the allocation
    SetAllocPadding((uint8*)allocMem,size);

    //retest max usage
    RecalcMaxMem();
}

//records freeing of memory
//returns the actual type of allocation (single or array, or none if invalid)
EMemAllocType MemMan::TraceFree(EMemAllocType type, volatile void *addr, const char *file, int line, char *&outAllocMem, int *outCount)
{
    if (!m_ready) //Framework not Initialized yet
    {
        BREAKPOINT;
        return EMAT_None;
    }
    
    if (outCount) *outCount=1;

    EMemAllocType realType=type;

    //check if they're deleting null memory first
    if (addr==0)
    {
        m_critErrors=true;

        MPMA::ErrorReport()<<"Error: delete being used on a null pointer.\n";
        MPMA::ErrorReport()<<"  -delete call made in "<<file<<" line "<<line<<".\n";
        MPMA::ErrorReport()<<"  -call stack:\n"<<MPMA::GetCallStack()<<"\n";

        return EMAT_None;
    }
    
    MPMA::TakeMutexLock autoSync(*syncLock);

    SAlloc *cur;
    SAlloc *prev=m_head;

    bool found=false;

    //see if theres even anything allocated
    if (m_head)
    {
        //find it in our list
        for (cur=m_head;cur;cur=cur->next)
        {
            if (cur->addr==addr) //found it    
            {
                found=true;
                
                //check that correct free method was used
                if (cur->type!=type)
                {
                    m_critErrors=true;
                    
                    MPMA::ErrorReport()<<"Error: Wrong delete operator used when freeing memory.\n";
                    
                    if (cur->type==EMAT_One) MPMA::ErrorReport()<<"  -Delete array used on memory allocated with new (non-array).\n";
                    else MPMA::ErrorReport()<<"  -Delete (non-array) used on memory allocated with new array.\n";
                    
                    MPMA::ErrorReport()<<"  -delete call made in "<<file<<" line "<<line<<".\n";
                    MPMA::ErrorReport()<<"  -call made on memory that was allocated in "<<cur->srcFile<<" line "<<cur->srcLine<<".\n";
                    MPMA::ErrorReport()<<"  -call stack:\n"<<MPMA::GetCallStack()<<"\n";

                    realType=cur->type;
                }

                //check padding around the allocation
                if (VerifyAllocPadding((uint8*)cur->allocMem, cur->size))
                {
                    m_critErrors=true;
                    MPMA::ErrorReport()<<"  -object: "<<cur->objName<<"\n";
                    MPMA::ErrorReport()<<"  -Allocated in "<<cur->srcFile<<" line "<<cur->srcLine<<".\n";
                    MPMA::ErrorReport()<<"  -Call stack for allocation:\n"<<cur->allocStack;
                    MPMA::ErrorReport()<<"  -Call stack for this delete:\n"<<MPMA::GetCallStack()<<"\n\n";
                }

                //output params
                if (outCount) *outCount=cur->count;
                outAllocMem=(char*)cur->allocMem;

                //if saving old ones, save a copy
                #ifdef MEM_TRACK_OLD_FREE
                    oldFreedAllocs.push_back(*cur);
                    oldFreedAllocs.back().next=0;
                    oldFreedAllocs.back().deleteStack=MISC::StripFirstLine(MPMA::GetCallStack());
                #endif
                
                //remove from list
                if (cur==m_head)
                {
                    m_head=m_head->next;
                    delete cur;
                    cur=0;
                }
                else
                {
                    prev->next=cur->next;
                    delete cur;
                    cur=0;
                }
                
                //
                break;
            }
            
            prev=cur;
        }
    }

    //if it wasn't found, they are deleting bad memory
    if (!found)
    {
        m_critErrors=true;

        MPMA::ErrorReport()<<"Error: delete used on memory that doesn't exist.\n";
        MPMA::ErrorReport()<<"  -delete call made in "<<file<<" line "<<line<<"\n";
        MPMA::ErrorReport()<<"  -call stack:\n"<<MPMA::GetCallStack()<<"\n";

        //check old records
        #ifdef MEM_TRACK_OLD_FREE
        bool oldFound=false;
        for (std::vector<MemMan::SAlloc>::iterator i=oldFreedAllocs.begin(); i!=oldFreedAllocs.end(); ++i)
        {
            if (i->addr==addr)
            {
                oldFound=true;
                MPMA::ErrorReport()<<"  -matching historic allocation found:\n";
                MPMA::ErrorReport()<<"   >object name: "<<i->objName<<"\n";
                MPMA::ErrorReport()<<"   >object size: "<<i->size<<"\n";
                MPMA::ErrorReport()<<"   >object's allocation stack:\n"<<i->allocStack;
                MPMA::ErrorReport()<<"   >object's delete stack:\n"<<i->deleteStack;
            }
        }
        if (!oldFound)
        {
            MPMA::ErrorReport()<<"  -No historic allocations found for the same address.  Likely a bad pointer passed.\n";
        }
        #endif
        MPMA::ErrorReport()<<"\n";

        return EMAT_None;
    }

    return realType;
}

//recalcs the max memory used
void MemMan::RecalcMaxMem()
{
#ifdef TRACK_MAX_MEM
    //calc current usage from allocs
    int curMem=0;

    SAlloc *cur=m_head;
    while (cur)
    {
        curMem+=cur->size;

        cur=cur->next;
    }

    //calc current usage from hints
    for (std::list<Hint>::iterator i=hints.begin(); i!=hints.end(); ++i)
    {
        curMem+=(*i).bytes;
    }

    //compare to max usage
    if (curMem>maxMem) maxMem=curMem;
#endif
}

//tells the memory manager that some memory has been allocated externally (used for tracking max usage only)
//returns an ID of the hinted allocation
int MemMan::HintMemAlloc(int bytes)
{
#ifdef TRACK_MAX_MEM
    //make a pair for the hint list and advance ID
    Hint newHint;
    newHint.ID=curHintAllocID;
    newHint.bytes=bytes;
    newHint.stack=MISC::StripFirstLine(MPMA::GetCallStack());

    curHintAllocID++;

    //add it to list
    hints.push_back(newHint);

    //recalc max usage
    RecalcMaxMem();

    //give back ID of the hint
    return newHint.ID;
#else
    return 0;
#endif
}

//tells the memory manager that some memory has been freed externally that was previously hinted at
void MemMan::HintMemFree(int ID)
{
#ifdef TRACK_MAX_MEM
    //remove specific ID from the list
    for (std::list<Hint>::iterator i=hints.begin(); i!=hints.end(); ++i)
    {
        if ((*i).ID==ID)
        {
            hints.erase(i);
            break;
        }
    }
#endif
}

//checks for memory corruption around all allocated memory
void MemMan::CheckForPaddingCorruption()
{
    MPMA::TakeMutexLock autoSync(*syncLock);
    
    for (SAlloc *cur=m_head;cur;cur=cur->next)
    {
        if (VerifyAllocPadding((uint8*)cur->allocMem, cur->size))
        {
            m_critErrors=true;
            std::string error="Memory corruption detected.\n";

            error+="  -object: "+cur->objName+"\n";
            error+="  -Allocated in "+cur->srcFile+" line "+Vary(cur->srcLine)+".\n";
            error+="  -Call stack for allocation:\n"+cur->allocStack+"\n";
            ReportLeak(error);
        }
    }
}

//verifies that a type and object declaration are the same type
void dbgN2_PrecheckType(const char *objName, const char *typName)
{
    int lenTyp=(int)strlen(typName);
    bool bad=false;
    if ((int)strlen(typName)>lenTyp) bad=true;

    if (!bad)
    {
        for (int i=0;i<lenTyp;++i)
            if (objName[i]!=typName[i]) bad=true;
    }

    if (bad)
    {
        MPMA::ErrorReport()<<"object declaration doesn't match type name provided to new2.\n";
        MPMA::ErrorReport()<<"  -type name: "<<typName<<"\n";
        MPMA::ErrorReport()<<"  -object dec: "<<objName<<"\n";
        MPMA::ErrorReport()<<"  -call stack:\n"<<MPMA::GetCallStack()<<"\n";
    }
}

#endif //MEMMAN_TRACING

//used by the allocation scheme, do not call directly - does a threadsafe alloc or free
char* MemMan::InternalAllocBlock(uint bytes)
{
#ifdef NEEDS_THREADSAFE_ALLOC
    if (!m_ready)
    {
        MPMA::ErrorReport()<<"MemMan being used before it is ready (will make a non-thread-safe allocation anyways here though)!\n";
        BREAKPOINT;
        return new char[bytes];
    }

    MPMA::TakeMutexLock autoSync(*syncLock);
#endif
    return new char[bytes];
}
void MemMan::InternalFreeBlock(volatile char *mem)
{
#ifdef NEEDS_THREADSAFE_ALLOC
    if (!m_ready)
    {
        MPMA::ErrorReport()<<"MemMan being used before it is ready!\n";
        BREAKPOINT;
        return;
    }

    MPMA::TakeMutexLock autoSync(*syncLock);
#endif
    delete[] mem;
}
