//!\file UDP.h UDP packet flinging
//written by Luke Lenhart, 2008
//See /docs/License.txt for details on how this code may be used.

#include "UDP.h"
#include "../base/MemMan.h"
#include "../base/DebugRouter.h"
#include "PlatformSockets.h"

namespace NET
{
    // -- receiver

    UDPReceiver::UDPReceiver()
    {
        sock=-1;
        port=0;
    }

    //Binds a socket to a port in order to receive upd packets.  If the port cannot be bound or an error occurs, returns 0.
    UDPReceiver* UDPReceiver::CreateReceiver(uint16 port)
    {
        //set us up the socket
        UDPReceiver *rec=new2(UDPReceiver,UDPReceiver);
        rec->port=port;
        rec->sock=socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
        if (rec->sock<0)
        {
            MPMA::ErrorReport()<<"Failed to create socket for udp receive.\n";
            FreeReceiver(rec);
            return 0;
        }

        //set nonblocking
        if (!Internal_SetSocketBlocking(rec->sock, false))
        {
            MPMA::ErrorReport()<<"Failed to set socket nonblocking mode.\n";
            //ignore it... better than just failing.
        }

        //bind it
        sockaddr_in addr;
        memset(&addr, 0, sizeof(addr));
        addr.sin_family=PF_INET;
        addr.sin_addr.s_addr=INADDR_ANY;
        addr.sin_port=htons(port);

        if (bind(rec->sock, (sockaddr*)&addr, sizeof(addr)))
        {
            MPMA::ErrorReport()<<"Failed to bind to port "<<port<<": "<<Internal_GetLastError()<<"\n";
            FreeReceiver(rec);
            return 0;
        }

        return rec;
    }

    //Unbinds and frees a UDPReceiver.
    void UDPReceiver::FreeReceiver(UDPReceiver *receiver)
    {
        if (receiver!=0)
        {
            Internal_CloseSocket(receiver->sock);
            delete2(receiver);
        }
    }

    //Receives the next packet and appends it to data.  source (if not 0) will be filled with the address the packet came from.  Returns false if there was nothing received.
    bool UDPReceiver::Receive(std::vector<uint8> &data, Address *source)
    {
        uint8 buf[64*1024]; //big enough to hold the max udp datagram size

        int count;
        if (!source)
            count=recv(sock, (char*)&buf[0], sizeof(buf), 0);
        else
        {
            sockaddr_in saddr;
            socklen_t saddrlen=sizeof(saddr);
            count=recvfrom(sock, (char*)&buf[0], sizeof(buf), 0, (sockaddr*)&saddr, &saddrlen);
            if (count>0)
            {
                source->SetAddress(inet_ntoa(saddr.sin_addr));
                source->SetPort(ntohs(saddr.sin_port));
            }
        }

        if (count<=0)
            return false;

        data.insert(data.end(), &buf[0], &buf[count]);
        return true;
    }

    // -- sender

    UDPSender::UDPSender()
    {
        sock=-1;
    }

    //Creates a socket to be used to send upd packet.  Returns 0 on failure.
    UDPSender* UDPSender::CreateSender()
    {
        //set us up the socket
        UDPSender *snd=new2(UDPSender,UDPSender);
        snd->sock=socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
        if (snd->sock<0)
        {
            MPMA::ErrorReport()<<"Failed to create socket for udp send.\n";
            FreeSender(snd);
            return 0;
        }

        return snd;
    }

    //Frees a UDPSender.
    void UDPSender::FreeSender(UDPSender *sender)
    {
        if (sender!=0)
        {
            Internal_CloseSocket(sender->sock);
            delete2(sender);
        }
    }

    //Sends a packet of data.  Returns false if the send failed.
    bool UDPSender::Send(const Address &addr, const void *data, uint dataLen)
    {
        if (dataLen>65507)
        {
            MPMA::ErrorReport()<<"Sending a packet of size "<<dataLen<<" which is greater than max udp packet size of "<<65507<<".  This will probably fail or be truncated.\n";
        }

        sockaddr_in saddr;
        saddr.sin_family=AF_INET;
        saddr.sin_addr.s_addr=inet_addr(addr.GetIPString().c_str());
        saddr.sin_port=htons(addr.GetPort());

        int count=sendto(sock, (const char*)data, dataLen, 0, (sockaddr*)&saddr, sizeof(saddr));
        if (count<=0)
        {
            MPMA::ErrorReport()<<"UDP send failed: "<<Internal_GetLastError()<<"\n";
            return false;
        }

        return true;
    }
}
